/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: adi2probe.c,v 1.1.1.1 1997/09/17 20:39:29 gropp Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpid.h"
#include "mpiddev.h"
#include "mpid_debug.h"
#include "queue.h"

void MPID_Iprobe(
	struct MPIR_COMMUNICATOR *comm_ptr, int tag, int context_id,
	int src_lrank, int *found, int *error_code, MPI_Status *status
) {
	MPIR_RHANDLE *rhandle;
	int position;
	MPID_Device *dev = MPID_devset->dev_list;

	DEBUG_PRINT_MSG("Entering Iprobe");
	DEBUG_PRINT_ARGS("Iprobe");

	/*
	 *  This is necessary as we might have disabled IRQ-driven
	 *  execution of GAMMA protocol,so we must try and extract
	 *  a possible pending msg explicitly.
	 */
	(*dev->check_device)(dev,MPID_NOTBLOCKING);

	/*  Check to see if the message has already been received.  */
	/*  Will not remove the found entry (flag = 0); */

	MPID_Atomic_Search_unexpected_queue(
		src_lrank, tag, context_id, 0, &rhandle, &position
	);

	if (rhandle != NULL) {
		if ((void *)rhandle == (void *)(-1)) {
			/*
			Go extract messages from the device to complete this
			entry of unexpected queue.
			*/
			do
			*error_code = (*dev->check_device)(dev,MPID_BLOCKING);
			while (!(MPID_Unexpected_entry_is_complete(position)));
			rhandle = MPID_Get_unexpected_entry(position);
		}
		*found  = 1;
		*status = rhandle->s;
		DEBUG_PRINT_MSG(" Iprobe found msg");
		DEBUG_PRINT_MSG("Exiting Iprobe");
		return;
	}

	/*
	 *  If nothing there, even incomplete, then nothing has arrived yet.
	 */
	*found = 0;
	DEBUG_PRINT_MSG(" Iprobe did not find msg");
	return;
}


void MPID_Probe(
	struct MPIR_COMMUNICATOR *comm_ptr, int tag, int context_id,
	int src_lrank, int *error_code, MPI_Status *status
) {
	int found;

	*error_code = MPI_SUCCESS;
	DEBUG_PRINT_MSG("Entering Probe");
	while (1) {
		MPID_Iprobe(
			comm_ptr, tag, context_id, src_lrank, &found,
			error_code, status
		);
		if (found || *error_code) {
			DEBUG_PRINT_MSG("Exiting Probe");
			return;
		}
		/* Wait for a message */
		MPID_DeviceCheck( MPID_BLOCKING );
	}
}
