/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: adi2req.c,v 1.2 1998/03/13 22:32:20 gropp Exp $
 *
 *  (C) 1996 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpid.h"
#include "mpiddev.h"
#include "reqalloc.h"
#include "sendq.h"	/* For MPIR_FORGET_SEND */


/* Free not completed requests. Tricky when request is for a non-local op. */

void MPID_Request_free (MPI_Request request) {
	int mpi_errno = MPI_SUCCESS;
    
	switch (request->handle_type) {
		case MPIR_SEND:
			if (MPID_SendIcomplete(request,&mpi_errno)) {
				MPIR_FORGET_SEND( &request->shandle );
				MPID_SendFree( request );
				request = 0;
			}
			break;
		case MPIR_RECV:
			if (
			MPID_RecvIcomplete(request,(MPI_Status *)0,&mpi_errno)
			){
				MPID_RecvFree( request );
				request = 0;
			}
			break;	
		case MPIR_PERSISTENT_SEND:
			MPID_Abort(
			(struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
			"Unimplemented operation - persistent send free"
			);
			break;
		case MPIR_PERSISTENT_RECV:
			MPID_Abort(
			(struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
			"Unimplemented operation - persistent recv free"
			);
			break;
	}

	/* 
	 * If we couldn't complete it, decrement it's reference count
	 * and forget about it.  This requires that the device detect
	 * orphaned requests when they do complete, and process them
	 * independent of any wait/test.
	 */
	if (request)
		request->chandle.ref_count--;
			/* (count == 0) --> "I'm orphan" */
	return;
}
