/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: chbrndv.c,v 1.2 1998/01/29 14:25:30 gropp Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


/*  ADI protocol:  Blocking Rendezvous  */


#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "reqalloc.h"
#include "chgamma.h"


/* Prototype definitions */

static int MPID_CH_Rndvb_ok_to_send
	ANSI_ARGS(( MPID_Aint, int ));
int MPID_CH_Rndvb_send
	ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t ));
int MPID_CH_Rndvb_isend
	ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t,
	MPIR_SHANDLE * ));
int MPID_CH_Rndvb_irecv
	ANSI_ARGS(( MPIR_RHANDLE *, int, void * ));
int MPID_CH_Rndvb_save
	ANSI_ARGS(( MPIR_RHANDLE *, int, void *));
int MPID_CH_Rndvb_unxrecv_start
	ANSI_ARGS(( MPIR_RHANDLE *, void * ));
int MPID_CH_Rndvb_ack
	ANSI_ARGS(( void *, int ));
void MPID_CH_Rndvb_delete
	ANSI_ARGS(( MPID_Protocol * ));
#if defined(MPID_RNDV_SELF)
static int MPID_CH_Rndvb_save_self
	ANSI_ARGS(( MPIR_RHANDLE *, int, void *));
static int MPID_CH_Rndvb_unxrecv_start_self
	ANSI_ARGS(( MPIR_RHANDLE *, void * ));
#endif


/* Globals for this protocol */

/*
 * Notes
 * In the case of sending a rendezvous message to self (source==destination),
 * there can be problems because the code expects to be able to send a 
 * request and then receive the requested data.  The sequence
 *  Send rendezvous (to self)
 *  Receive as unexpected
 *  Post receive and send "ok to send" (to self)
 *  Wait by entering blocking receive for message (from self)
 * can fail if the process does not receive its ok "ok to send" message.
 * We can fix this in two ways
 *  Add a DeviceCheck before trying to complete the rendezvous.
 *  Use a different set of routines for handling sends to self in the
 *   rendezvous case.
 * While the second case looks like the obvious thing to do, one problem
 * with it is that some systems provide better self-to-self copy when using
 * their communication network.  We could take the position that these systems
 * have badly designed memory systems, but it could be a problem.  
 * See aditest12 for an example.
 */


/*
 *  Definitions of the protocol functions
 */


/*
 *  This GAMMA handler is only for ports of control messages
 *  (REQUEST_SEND, OK_TO_SEND).
 *  Data messages (both short-eager and rendez-vous protocols) are managed
 *  by the other handler (see chshort.c).
 */
void AGR_handler_rec(void) {

register unsigned long t1,t3,t4,peppe=0;

#ifndef AG_FLOWCTL
if (((AG_head + 1) % AG_MAXQLEN) != AG_tail) {
#endif

t1 = gamma_active_port-gamma_how_many_nodes();
AG_src_queue[AG_head].src = t1;

if (
((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->mode == MPID_PKT_REQUEST_SEND
) {

	/*  Request to send */

	/*  Memcpy msg header into the queue of incoming headers.  */
	memcpy(&AG_pkt_queue[AG_head].p,
		AGR_buffer[t1],sizeof(MPID_PKT_REQUEST_SEND_T));

	/*  Save the id. of this request.  */
	AG_src_queue[AG_head].temp_buf = (void *)((MPID_PKT_REQUEST_SEND_T *)
		AGR_buffer[t1])->send_id;

	/*
	 *  Is this expected or not?
	 *  Try to find a matching receive request in posted queue,
	 *  excluding those ones already marked as "rndv_pending".
	 *  This is similar to MPID_Search_posted_queue() (queue.c).
	 */
	if (MPID_recvs.posted.how_many) {
		t4 = MPID_recvs.posted.head;
		t3 = t4;
		/*  Safety guard.  */
		for (;peppe < AG_MAXQLEN+10;peppe++) {
			if (
((((MPID_PKT_HEAD_T *)
(AGR_buffer[t1]))->context_id) == MPID_recvs.posted.queue[t4].context_id)
&&
((((((MPID_PKT_HEAD_T *)
(AGR_buffer[t1]))->tag)
^ MPID_recvs.posted.queue[t4].tag) & MPID_recvs.posted.queue[t4].tagmask) == 0)
&&
((((((MPID_PKT_HEAD_T *)
(AGR_buffer[t1]))->lrank)
^ MPID_recvs.posted.queue[t4].lsrc)& MPID_recvs.posted.queue[t4].srcmask) == 0)
&&
(MPID_recvs.posted.queue[t4].rndv_pending == 0)
			) {
				/*
				 *  Message is expected, but this is only
				 *  the REQUEST_SEND, not the delivery.
				 *  Therefore we MUST leave the recv in the
				 *  posted queue.
				 *  We also must mark it as "rndv_pending".
				 */
				MPID_recvs.posted.queue[t4].rndv_pending = 1;
				AG_src_queue[AG_head].unexp_entry = -1;
				AG_src_queue[AG_head].rhandle =
					MPID_recvs.posted.queue[t4].ptr;
				AG_head = (AG_head + 1) % AG_MAXQLEN;
				gamma_signal(AG_NEWMSG_SEM);
				return;
			}
			if (t4 == MPID_recvs.posted.tail) break;
			t3 = t4;
			t4 = MPID_recvs.posted.queue[t4].next;
		}
		if (peppe == AG_MAXQLEN+10) {
			/*  Oops!  Inconsistent recv queue!  */
			/* peppe qui perfezionare la segnalazione di errore */
			gamma_sigerr(AG_NEWMSG_SEM);
			return;
		}
	}

	/*  Message is unexpected.  */
	/*
	 *  Find the first free position of unexpected queue.
	 *  Then, pre-compute the entry by filling up all the
	 *  data relevant to a future receive match.
	 *  The entry is incomplete, since no rhandle
	 *  is allocated here; it will be
	 *  completed by MPID_CH_Check_incoming() (chchkdev.c).
	 *  This is similar to a subset of MPID_Enqueue() (queue.c).
	 */
#ifndef AG_FLOWCTL
	if (MPID_recvs.unexpected.how_many == AG_MAXQLEN) {	
		/* No room in unexpected queue. Message discarded. */
/* peppe qui perfezionare la segnalazione di errore */
		goto ERR;
	}
#endif

	for (t4 = MPID_recvs.unexpected.tail;;t4 = (t4+1) % AG_MAXQLEN)
		if (MPID_recvs.unexpected.queue[t4].free) break;
	/*  We now assume that the above loop provided a value for t4.  */
	if (MPID_recvs.unexpected.how_many) {
		MPID_recvs.unexpected.queue[MPID_recvs.unexpected.tail].next
		= t4;
		MPID_recvs.unexpected.tail = t4;
	}
	MPID_recvs.unexpected.how_many++;
	MPID_recvs.unexpected.queue[t4].free = 0;
	MPID_recvs.unexpected.queue[t4].next = 0;
	/* Store the selection criteria is an easy-to-get place */
	MPID_recvs.unexpected.queue[t4].context_id =
	((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->context_id;
	MPID_recvs.unexpected.queue[t4].tag =
	((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->tag;
	MPID_recvs.unexpected.queue[t4].lsrc =
	((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->lrank;
	MPID_recvs.unexpected.queue[t4].ptr = (MPIR_RHANDLE *)(-1);
	AG_src_queue[AG_head].unexp_entry = t4;
	AG_head = (AG_head + 1) % AG_MAXQLEN;
	gamma_signal(AG_NEWMSG_SEM);
	return;

}

if (
((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->mode == MPID_PKT_OK_TO_SEND
) {

	/*  OK_TO_SEND control message.  */

	/*  Memcpy msg header into the queue of incoming headers.  */
	memcpy(&AG_pkt_queue[AG_head].p,
		AGR_buffer[t1],sizeof(MPID_PKT_OK_TO_SEND_T));

	t3 = (unsigned long)((MPID_PKT_OK_TO_SEND_T *)AGR_buffer[t1])->send_id;
	if ( ((MPIR_SHANDLE *)t3)->ref_count > 0 ) {
		/*  It will be processed by MPID_CH_Check_incoming().  */
		/*  Mark this control msg as "expected"  */
		AG_src_queue[AG_head].unexp_entry = -1;
		AG_head = (AG_head + 1) % AG_MAXQLEN;
		gamma_signal(AG_NEWMSG_SEM);
		return;
	}

	/*  Oops! The related send request has been freed meanwhile!!  */
	/*  Try to start from ourselves... */
	AG_pkt.mode = MPID_PKT_LONG;
	AG_pkt.lrank = ((MPIR_SHANDLE *)t3)->src_lrank;
	AG_pkt.context_id = ((MPIR_SHANDLE *)t3)->context_id;
	AG_pkt.tag = ((MPIR_SHANDLE *)t3)->tag;
	AG_pkt.len = ((MPIR_SHANDLE *)t3)->bytes_as_contig;
	MPID_DO_HETERO(AG_pkt.msgrep = (int)((MPIR_SHANDLE *)t3)->msgrep);
	MPID_PKT_PACK(
		&AG_pkt, sizeof(MPID_PKT_HEAD_T),((MPIR_SHANDLE *)t3)->dest);

	/*
	 *  We cannot use flow-controlled GAMMA sends here, because they
	 *  require to poll the NIC for incoming flow credits, and this is
	 *  impossible to do correctly from inside GAMMA receive handlers.
	 *  Therefore we violate the flow control model in this case.
	 *  Please note that this is a very particular and hopefully unfrequent
	 *  case: the receiver handler is trying to finish a rendez-vous
	 *  communication whose send request has been withdrawn meanwhile.
	 */
	GAMMASEND_2P_FOR_HANDLER(
		((MPIR_SHANDLE *)t3)->dest,
		&AG_pkt,sizeof(MPID_PKT_HEAD_T),
		((MPIR_SHANDLE *)t3)->start,
		((MPIR_SHANDLE *)t3)->bytes_as_contig
	);

	/*  Operations may be finished by MPID_CH_Check_incoming().  */
	/*  Mark this control msg as "unexpected"  */
	AG_src_queue[AG_head].unexp_entry = 0;
	AG_head = (AG_head + 1) % AG_MAXQLEN;
	gamma_signal(AG_NEWMSG_SEM);
	return;

}

/*  Should NEVER arrive here!!  */
if (((MPID_PKT_HEAD_T *)(AGR_buffer[t1]))->len) {
	/*  Discard payload.  */
	gamma_switch_to_buffer(t1,NULL,0);
}
AG_src_queue[AG_head].unexp_entry = -1;
AG_head = (AG_head + 1) % AG_MAXQLEN;
gamma_sigerr(AG_NEWMSG_SEM);
return;

#ifndef AG_FLOWCTL
}
/*
 *  Next ctl/short msg would overflow the AG_src_queue.
 *  It is much easier to detect the overflow now.
 *  So we assume that there is overflow, and discard
 *  this message by not notifying it.
 */
/* peppe qui perfezionare la segnalazione di errore */
ERR:
gamma_sigerr(AG_NEWMSG_SEM);
return;
#endif

}


/* Internal routine to send an OK_TO_SEND */

static int MPID_CH_Rndvb_ok_to_send(MPID_Aint send_id, int from)
{
	MPID_PKT_OK_TO_SEND_T pkt;

	pkt.mode = MPID_PKT_OK_TO_SEND;
	MPID_AINT_SET(pkt.send_id,send_id);
	DEBUG_PRINT_BASIC_SEND_PKT("S Ok send", &pkt);
	MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_OK_TO_SEND_T), from );
#if AGR_DEBUG
	printf("[%d] irecv: sending ok_to_send, from %d, id %d\n",
		gamma_my_node(),from,(int)send_id);
	MPID_Dump_queues();
	fflush(stdout);
#endif

	GAMMASEND(from+gamma_how_many_nodes(), &pkt,
			sizeof(MPID_PKT_OK_TO_SEND_T));

	return MPI_SUCCESS;
}


/* Internal routine to send a long message with rendezvous. */

static int MPID_CH_Eagerb_send_long_noflowctl(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep
) {
	MPID_PKT_HEAD_T pkt;

	/*
	 *  These references are ordered to match the order they appear in the 
	 *  structure.
	 */
	pkt.mode = MPID_PKT_LONG;
	pkt.lrank = src_lrank;
	pkt.context_id = context_id;
	pkt.tag = tag;
	pkt.len = len;
	MPID_DO_HETERO(pkt.msgrep = (int)msgrep);
	DEBUG_PRINT_SEND_PKT("S Sending",&pkt);
	MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_HEAD_T), dest );
	DEBUG_PRINT_PKT_DATA("S Getting data from buf",&pkt);
	DEBUG_PRINT_SEND_PKT("S Sending long msg",&pkt);

#if AGR_DEBUG
	printf("[%d] sending %d bytes to port %d\n",gamma_my_node(),len,dest);
	fflush(stdout);
#endif
	GAMMASEND_2P(dest,&pkt,sizeof(MPID_PKT_HEAD_T),buf,len);

	DEBUG_PRINT_MSG("S Sent long msg");
	return MPI_SUCCESS;
}


/*  Blocking send  */

static int Rndvb_isend_special(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep, MPIR_SHANDLE *shandle
) {
	MPID_PKT_REQUEST_SEND_T pkt;

	/* Store info in the request for completing the send later */
	shandle->is_complete = 0;
	shandle->start = buf;
	shandle->bytes_as_contig = len;
	shandle->src_lrank = src_lrank;
	shandle->tag = tag;
	shandle->context_id = context_id;
	shandle->dest = dest;
	shandle->msgrep = msgrep;
	/* Set the test/wait functions */
	shandle->wait = MPID_WaitForCompleteSend;
	/*  Send the REQUEST SEND message.  */
	DEBUG_PRINT_BASIC_SEND_PKT("S Sending rndv message",&pkt)
	pkt.mode = MPID_PKT_REQUEST_SEND;
	pkt.lrank = src_lrank;
	pkt.context_id = context_id;
	pkt.tag = tag;
	pkt.len = len;
	MPID_AINT_SET(pkt.send_id,shandle);
	MPID_DO_HETERO(pkt.msgrep = (int)msgrep);
	MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_REQUEST_SEND_T), dest );
	MPID_n_pending++;
#if AGR_DEBUG
printf("[%d] sending req to port %d: mode %d, start %p, len %d, lrank %d, context %d, tag %d, len %d, id %ld\n",
gamma_my_node(),
dest+gamma_how_many_nodes(),
pkt.mode,
buf,len,
pkt.lrank,
pkt.context_id,
pkt.tag,
pkt.len,
(unsigned long)pkt.send_id);
fflush(stdout);
#endif
	if (dest != MPID_MyWorldRank) {
#ifdef AG_FLOWCTL
	if (AG_flow[dest].msgs) {
		AG_flow[dest].msgs--;
#endif
		/*  Send a REQUEST_SEND ctl message.  */
		GAMMASEND(dest+gamma_how_many_nodes(),
			&pkt,sizeof(MPID_PKT_REQUEST_SEND_T));
		return MPI_SUCCESS;
#ifdef AG_FLOWCTL
	}
	/*  message credit towards dest exhausted. Shouldn't occur!  */
	fprintf(stderr,
	"[%d] Rndvb_isend_special(): Msg credit towards #%d exhausted!!!\n",
		gamma_my_node(),dest);
	return MPI_ERR_EXHAUSTED;
#endif
	}

	/*  ``Send to self''  */
	{
	MPIR_RHANDLE *rhandle = NULL;
	MPID_Atomic_Search_posted_queue(dest,tag,context_id,1,&rhandle);
	if (rhandle != NULL) {
		/*  Was expected.  */
		MPID_CH_Rndvb_irecv(rhandle, dest, &pkt);
	}
	else {
		/*  Was unexpected!  */
		/*  Prepare temporary rhandle...  */
		MPID_RecvAlloc(rhandle);
		if (!rhandle) {
			MPIR_ERROR(
			MPIR_COMM_WORLD, MPI_ERR_INTERN,
			"Could not dynamically allocate internal handle"
			);
			return MPI_ERR_INTERN;
		}
		MPID_RecvInit(rhandle);

		/*  Insert into unexpected queue...  */
		MPID_Atomic_Unexp(dest,tag,context_id,NULL,len,rhandle);

		/*  Call the save routine.  */
		MPID_CH_Rndvb_save_self(rhandle, dest, &pkt);
	}
	return MPI_SUCCESS;
	}
}

int MPID_CH_Rndvb_send(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep
) {
	MPIR_SHANDLE shandle;
	int ret;

	DEBUG_INIT_STRUCT(&shandle,sizeof(shandle));
	MPIR_SET_COOKIE((&shandle),MPIR_REQUEST_COOKIE);
	MPID_SendInit( &shandle );
	ret = Rndvb_isend_special(
		buf, len, src_lrank, tag, context_id, dest, msgrep, &shandle);
	if (ret != MPI_SUCCESS)
		return ret;
	DEBUG_TEST_FCN(shandle.wait,"req->wait");
	shandle.wait(&shandle);
	return MPI_SUCCESS;
}


/*  Non-blocking send  */

int MPID_CH_Rndvb_isend(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep, MPIR_SHANDLE *shandle
) {
	MPID_PKT_REQUEST_SEND_T pkt;

	/* Store info in the request for completing the send later */
	shandle->is_complete = 0;
	shandle->start = buf;
	shandle->bytes_as_contig = len;
	shandle->src_lrank = src_lrank;
	shandle->tag = tag;
	shandle->context_id = context_id;
	shandle->dest = dest;
	shandle->msgrep = msgrep;
	/* Set the test/wait functions */
	shandle->wait = MPID_WaitForCompleteSend;
	/*  Send the REQUEST SEND message.  */
	DEBUG_PRINT_BASIC_SEND_PKT("S Sending rndv message",&pkt)
	pkt.mode = MPID_PKT_REQUEST_SEND;
	pkt.lrank = src_lrank;
	pkt.context_id = context_id;
	pkt.tag = tag;
	pkt.len = len;
	MPID_AINT_SET(pkt.send_id,shandle);
	MPID_DO_HETERO(pkt.msgrep = (int)msgrep);
	MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_REQUEST_SEND_T), dest );
	MPID_n_pending++;
	if (dest != MPID_MyWorldRank) {
#if AGR_DEBUG
printf("[%d] sending req to port %d: mode %d, start %p, len %d, lrank %d, context %d, tag %d, len %d, id %ld\n",
gamma_my_node(),
dest+gamma_how_many_nodes(),
pkt.mode,
buf,len,
pkt.lrank,
pkt.context_id,
pkt.tag,
pkt.len,
(unsigned long)pkt.send_id);
fflush(stdout);
#endif
#ifdef AG_FLOWCTL
	/*  check if we still have msg credits left  */
	if (AG_flow[dest].msgs) {
		AG_flow[dest].msgs--;
#endif
		/*  Send a REQUEST_SEND ctl message.  */
		GAMMASEND(dest+gamma_how_many_nodes(),
			&pkt,sizeof(MPID_PKT_REQUEST_SEND_T));
		return MPI_SUCCESS;
#ifdef AG_FLOWCTL
	}
	/*  message credit towards "dest" exhausted.  Failure.  */
	fprintf(stderr,
	"[%d] MPID_CH_Rndvb_isend(): Msg credit towards #%d exhausted\n",
		gamma_my_node(),dest);
	return MPI_ERR_EXHAUSTED;
#endif
	}

	/*  ``Send to self''  */
	{
	MPIR_RHANDLE *rhandle = NULL;
	MPID_Atomic_Search_posted_queue(src_lrank,tag,context_id,1,&rhandle);
	if (rhandle != NULL) {
		/*  Was expected.  */
		MPID_CH_Rndvb_irecv(rhandle, dest, &pkt);
	}
	else {
		/*  Was unexpected!  */
		/*  Prepare temporary rhandle...  */
		MPID_RecvAlloc(rhandle);
		if (!rhandle) {
			MPIR_ERROR(
			MPIR_COMM_WORLD, MPI_ERR_INTERN,
			"Could not dynamically allocate internal handle"
			);
			return MPI_ERR_INTERN;
		}
		MPID_RecvInit(rhandle);

		/*  Insert into unexpected queue...  */
		MPID_Atomic_Unexp(src_lrank,tag,context_id,NULL,len,rhandle);

		/*  Call the save routine.  */
		MPID_CH_Rndvb_save_self(rhandle, dest, &pkt);
	}
	return MPI_SUCCESS;
	}
}


/* This is called when an expected REQUEST SEND is seen. */

int MPID_CH_Rndvb_irecv(MPIR_RHANDLE *rhandle, int from, void *in_pkt)
{
	MPID_PKT_REQUEST_SEND_T *pkt = (MPID_PKT_REQUEST_SEND_T *)in_pkt;
	int err = MPI_SUCCESS;
#if defined(MPID_RNDV_SELF)
	MPIR_SHANDLE *shandle;
#endif

	DEBUG_PRINT_MSG("R Starting rndvb irecv");

	/* A request packet is a little larger than the basic packet size and 
	may need to be unpacked (in the heterogeneous case) */
	MPID_PKT_UNPACK(
		(MPID_PKT_HEAD_T *)in_pkt + 1, 
		sizeof(MPID_PKT_REQUEST_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		from
	);

	/* Check for truncation */
	MPID_CHK_MSGLEN(rhandle,pkt->len,err,pkt)
	/* Note that if we truncate, We really must receive the message in two 
	parts; the part that we can store, and the part that we discard.
	This case is not yet handled. */
	rhandle->s.count = pkt->len;
	rhandle->s.MPI_TAG = pkt->tag;
	rhandle->s.MPI_SOURCE = pkt->lrank;
	rhandle->s.MPI_ERROR = err;
#if defined(MPID_RNDV_SELF)
	if (from == MPID_MyWorldRank) {
		DEBUG_PRINT_MSG("R Starting a receive transfer from self");
		MPID_AINT_GET(shandle,pkt->send_id);
#ifdef MPIR_HAS_COOKIES
		if (shandle->cookie != MPIR_REQUEST_COOKIE) {
			fprintf( stderr, "shandle is %lx\n", (long)shandle );
			fprintf( stderr, "shandle cookie is %lx\n",
				shandle->cookie );
			MPID_Print_shandle( stderr, shandle );
			MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1,
				"MPI internal", 
			"Bad address in Rendezvous send (irecv-self)" );
		}
#endif	
		/* Copy directly from the shandle */
		MEMCPY(rhandle->buf,shandle->start,shandle->bytes_as_contig);

		/*  Complete the self send operation.  */
		shandle->is_complete = 1;
		MPID_n_pending--;

		/* Update all of the rhandle information */
/* peppe forse superflue o forse necessarie per le persistent ops  */
#if 0
		rhandle->push	 = 0;
#endif

		/*  Complete the self recv operation.  */
		rhandle->is_complete = 1;
		if (rhandle->finish)
			(rhandle->finish)( rhandle );
		return err;
	}
#endif
	rhandle->send_id = (MPID_Aint)AGR_send_id;

	/*  Send OK_TO_SEND back to the sender.  */
	MPID_CH_Rndvb_ok_to_send(rhandle->send_id, from);

	return err;
}


/* This is called when an OK TO SEND is received. */

int MPID_CH_Rndvb_ack(void *in_pkt, int from_grank) {
	MPID_PKT_OK_TO_SEND_T *pkt = (MPID_PKT_OK_TO_SEND_T *)in_pkt;
	MPIR_SHANDLE *shandle = NULL;

	MPID_AINT_GET(shandle,pkt->send_id);
	DEBUG_PRINT_MSG("Sending data on channel");
	MPID_n_pending--;
#ifdef AG_FLOWCTL
	/*  Refill of message credit, through rendez-vous  */
	AG_flow[from_grank].msgs += AG_MAXQLEN/(2*(MPID_MyWorldSize-1));
#endif
	/*
	 *  Use the eager protocol to send data:
	 *  We are now sure that a matching recv is posted on the recv side.
	 */
	MPID_CH_Eagerb_send_long_noflowctl(
		shandle->start,
		shandle->bytes_as_contig,
		shandle->src_lrank,
		shandle->tag,
		shandle->context_id,
		shandle->dest,
		shandle->msgrep
	);
	DEBUG_PRINT_MSG("Completed sending data on channel");
	shandle->is_complete = 1;
	if (shandle->ref_count == 0)
		MPID_SendFree((MPI_Request *)shandle);
	return MPI_SUCCESS;
}


/* This is called when an OK TO SEND is received and the request was freed. */

int MPID_CH_Rndvb_ack_freed(void *in_pkt, int from_grank) {
	MPID_PKT_OK_TO_SEND_T *pkt = (MPID_PKT_OK_TO_SEND_T *)in_pkt;
	MPIR_SHANDLE *shandle = NULL;

	MPID_AINT_GET(shandle,pkt->send_id);
	MPID_n_pending--;
	MPID_SendFree((MPI_Request *)shandle);
	return MPI_SUCCESS;
}


/* Save an unexpected REQUEST_SEND in rhandle */

int MPID_CH_Rndvb_save(MPIR_RHANDLE *rhandle, int from, void *in_pkt)
{
	MPID_PKT_HEAD_T *pkt = (MPID_PKT_HEAD_T *)in_pkt;

	rhandle->s.count = pkt->len;
	rhandle->s.MPI_ERROR = 0;
	rhandle->send_id = (MPID_Aint)AGR_send_id;
	rhandle->from = from;
	MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt->msgrep);

	/* Need to set the push routine to complete this transfer */
	rhandle->push = MPID_CH_Rndvb_unxrecv_start;
	return 0;
}


/*  This is called when it is time to receive an unexpected message.  */

int MPID_CH_Rndvb_unxrecv_start(
	MPIR_RHANDLE *rhandle, void *in_runex
) {
	MPIR_RHANDLE *runex = (MPIR_RHANDLE *)in_runex;

	rhandle->s = runex->s;

	/*  Post the recv in order to catch the msg when it arrives.  */
	MPID_Atomic_Post_rndv(
		rhandle->s.MPI_SOURCE,
		rhandle->s.MPI_TAG,
		rhandle->contextid,
		rhandle->buf,
		rhandle->len,
		rhandle
	);

	/*  Send OK_TO_SEND back to the sender.  */
	MPID_CH_Rndvb_ok_to_send(runex->send_id, runex->from);

	MPID_RecvFree( runex );
	return 0;
}


#if defined(MPID_RNDV_SELF)
/************************************************************************
 * This code handles rendez-vous messages to self.			*
 ************************************************************************/

/* Save an unexpected message in rhandle for sent to self */

static int MPID_CH_Rndvb_save_self(
	MPIR_RHANDLE *rhandle, int from, void *in_pkt
) {
	MPID_PKT_REQUEST_SEND_T   *pkt = (MPID_PKT_REQUEST_SEND_T *)in_pkt;

	/* A request packet is a little larger than the basic packet size and 
	may need to be unpacked (in the heterogeneous case) */
	MPID_PKT_UNPACK(
		(MPID_PKT_HEAD_T *)in_pkt + 1, 
		sizeof(MPID_PKT_REQUEST_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		from
	);

	rhandle->s.MPI_TAG    = pkt->tag;
	rhandle->s.MPI_SOURCE = pkt->lrank;
	rhandle->s.MPI_ERROR  = 0;
	rhandle->s.count      = pkt->len;
	rhandle->is_complete  = 0;
	rhandle->from         = from;
	rhandle->send_id      = pkt->send_id;
	/* Note that the send_id is just the address of the sending handle
	IN OUR ADDRESS SPACE */

	/* Note that format MIGHT differ from native (e.g., packed data) */
	MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt->msgrep );

	/* Need to set the push etc routine to complete this transfer */
	rhandle->push = MPID_CH_Rndvb_unxrecv_start_self;

	return 0;
}


/* 
 * This is called when it is time to receive an unexpected
 * message.  This is simple; we can just copy the data with memcpy.
 * Once the memcpy is done, we mark the SEND as completed.
 */

static int MPID_CH_Rndvb_unxrecv_start_self(
	MPIR_RHANDLE *rhandle, void *in_runex
) {
	MPIR_RHANDLE *runex = (MPIR_RHANDLE *)in_runex;
	MPIR_SHANDLE *shandle;

	/* Get the source handle */
	MPID_AINT_GET(shandle,runex->send_id);
#ifdef MPIR_HAS_COOKIES
	if (shandle->cookie != MPIR_REQUEST_COOKIE) {
		fprintf( stderr, "shandle is %lx\n", (long)shandle );
		fprintf( stderr, "shandle cookie is %lx\n", shandle->cookie );
		MPID_Print_shandle( stderr, shandle );
		MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
			"Bad address in Rendezvous send (unx_start_self)" );
	}
#endif	

	/* Copy payload directly from the shandle buffer */
	MEMCPY( rhandle->buf, shandle->start, shandle->bytes_as_contig );

	/*  Complete the self send operation.  */
	shandle->is_complete = 1;
	MPID_n_pending--;

	/* Update all of the rhandle information */
	rhandle->s	 = runex->s;
	rhandle->push	 = 0;
	rhandle->from    = runex->from;

	/*  Destroy the temporary rhandle.  */
	MPID_RecvFree( runex );

	/*  Complete the self unexpected recv operation.  */
	rhandle->is_complete = 1;
	return 0;
}

/***************************************************************************/
#endif /* MPID_RNDV_SELF */


/* 
 * This is fairly hard.  We need to send a "please_cancel_send", 
 * which, if the message is found in the unexpected queue, removes it.
 * However, if the message is being received at the "same" moment, the
 * ok_to_send and cancel_send messages could cross.  To handle this, the
 * receiver must ack the cancel_send message (making the success of the
 * cancel non-local).  There are even more complex protocols, but we won't
 * bother.
 * 
 * Don't forget to update MPID_n_pending as needed.
 */

void MPID_CH_Rndvb_delete( p )
MPID_Protocol *p;
{
	FREE( p );
}


MPID_Protocol *MPID_CH_Rndvb_setup()
{
	MPID_Protocol *p;

	p = (MPID_Protocol *) MALLOC( sizeof(MPID_Protocol) );
	if (!p) return 0;
	p->send		= MPID_CH_Rndvb_send;
	p->recv		= 0;
	p->isend	= MPID_CH_Rndvb_isend;
	p->push_send	= 0;
	p->cancel_send	= 0;
	p->irecv	= MPID_CH_Rndvb_irecv;
	p->push_recv	= 0;
	p->cancel_recv	= 0;
	p->do_ack	= MPID_CH_Rndvb_ack;
	p->unex		= MPID_CH_Rndvb_save;
	p->delete	= MPID_CH_Rndvb_delete;

	return p;
}
