/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

#include "sizes.h"
#include "memlock.h"
#include <linux/gamma/libgamma.h>
#include <stdio.h>


#ifdef AG_HARDMEMLOCK

#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))

#define AG_MEM_DEBUG(a)
#define AG_MEM_DEBUG_LIGHT(a)
#define AG_MEM_DEBUG_VERY_LIGHT(a)

#if 1
int AG_mlock(void *buf,unsigned long size) {
	return gamma_mlock(buf,size);
}
void AG_munlock(void *buf, unsigned long size) {
	gamma_munlock(buf,size);
}
#else
int AG_mlock(void *buf,unsigned long size) {
	int i,j,n,m;

	if (size == 0) return 0;

	n = AG_lock.how_many;
	if (n > 0) {
		for (	i = (AG_lock.head==0 ? AG_MAXLOCK-1 : AG_lock.head-1);
			n;
			n--, i = (i==0 ? AG_MAXLOCK-1 : i-1)
		) {
			if (
(((unsigned long)buf)
>= (unsigned long)(AG_lock.table[i].addr))
&&
(((unsigned long)buf+size)
<= ((unsigned long)(AG_lock.table[i].addr)+AG_lock.table[i].len))
			) {
				/*
				 *  Buffer inclusion with an already
				 *  locked buffer.
				 */
AG_MEM_DEBUG(
printf("[%d] AG_mlock(): already locked, how_many %d, pos %d, in_use %d, found after %d steps\n",
gamma_my_node(),
AG_lock.how_many,i,AG_lock.table[i].in_use,AG_lock.how_many-n+1);
fflush(stdout);
)
				AG_lock.table[i].in_use++;
				return 0;
			}
			if (
(
(((unsigned long)buf)
>= (unsigned long)(AG_lock.table[i].addr))
&&
(((unsigned long)buf)
<= ((unsigned long)(AG_lock.table[i].addr)+AG_lock.table[i].len))
)
||
(
(((unsigned long)buf+size)
>= (unsigned long)(AG_lock.table[i].addr))
&&
(((unsigned long)buf+size)
<= ((unsigned long)(AG_lock.table[i].addr)+AG_lock.table[i].len))
)
			) {
				/*
				 *  Buffer intersection with an already
				 *  locked buffer.
				 */
AG_MEM_DEBUG_LIGHT(
printf("[%d] AG_mlock(): lock intersection, how_many %d, pos %d, in_use %d\n",
gamma_my_node(), AG_lock.how_many,i,AG_lock.table[i].in_use);
fflush(stdout);
)
				AG_lock.table[i].in_use++;
		/*  "Enlarge" the currently locked buffer.  */
		AG_lock.table[i].addr = (void *)
		min((unsigned long)buf,(unsigned long)AG_lock.table[i].addr);
		AG_lock.table[i].len =
		max(
(unsigned long)buf+size,
(unsigned long)AG_lock.table[i].addr+AG_lock.table[i].len
		) - (unsigned long)AG_lock.table[i].addr;
		if (
			gamma_mlock(AG_lock.table[i].addr,AG_lock.table[i].len)
			>= 0
		) {
AG_MEM_DEBUG_LIGHT(
printf("[%d] AG_mlock(): lock enlargment OK\n",
gamma_my_node());
fflush(stdout);
)
			return 0;
		}

		/*
		Lockable memory runs low!
		Try unlocking
		all the unlockable buffers...
		*/
AG_MEM_DEBUG_LIGHT(
printf("[%d] AG_mlock(): Trying unlocking all the unlockable buffers...\n",
gamma_my_node());
fflush(stdout);
)
		for (j=0;j<AG_lock.how_many;j++)
			if (!(AG_lock.table[j].in_use)) {
				gamma_munlock(
					AG_lock.table[j].addr,
					AG_lock.table[j].len
				);
				AG_lock.table[j].addr = NULL;
				AG_lock.table[j].len = 0;
			}
		/*  Now retry locking...  */
		if (
			gamma_mlock(AG_lock.table[i].addr,AG_lock.table[i].len)
			>= 0
		) {
AG_MEM_DEBUG_LIGHT(
printf("[%d] AG_mlock(): lock enlargment OK\n",
gamma_my_node());
fflush(stdout);
)
			return 0;
		}

		/*  mlock failed!  Bad news...  */
printf("[%d] AG_mlock(): out of lockable RAM!!!\n",
gamma_my_node());
fflush(stdout);
		return -1;
			}
		}

		/*  Add a new entry in the table.  */

		if (AG_lock.how_many < AG_MAXLOCK)
			AG_lock.how_many++;
		else {
			/*  No room in lock table.  Must evict an entry...  */
AG_MEM_DEBUG_VERY_LIGHT(
printf("[%d] AG_mlock(): Searching a buffer to evict...\n",
gamma_my_node());
fflush(stdout);
)
			for (	m = AG_lock.head;
				m < AG_lock.head + AG_MAXLOCK;
				m++
			)
				if (!(AG_lock.table[m % AG_MAXLOCK].in_use))
					break;
			if (m != AG_lock.head+AG_MAXLOCK) {
				m = m % AG_MAXLOCK;
AG_MEM_DEBUG_VERY_LIGHT(
printf("[%d] AG_mlock(): Evicting entry %d\n",
gamma_my_node(),m);
fflush(stdout);
)
				AG_lock.head = m;
				gamma_munlock(
					AG_lock.table[m].addr,
					AG_lock.table[m].len
				);
				AG_lock.table[m].addr = NULL;
				AG_lock.table[m].len = 0;
			}
			else {
AG_MEM_DEBUG_VERY_LIGHT(
printf("[%d] AG_mlock(): could not evict any buffer from lock table!!!\n",
gamma_my_node());
fflush(stdout);
)
				return -1;
			}
		}

		AG_lock.table[AG_lock.head].addr = (buf);
		AG_lock.table[AG_lock.head].len = (size);
		AG_lock.table[AG_lock.head].in_use++;
AG_MEM_DEBUG_VERY_LIGHT(
printf("[%d] AG_mlock(): locking: how_many %d, head %d, in_use %d\n",
gamma_my_node(), AG_lock.how_many,
(AG_lock.head+1)%AG_MAXLOCK,AG_lock.table[AG_lock.head].in_use);
fflush(stdout);
)
		AG_lock.head = (AG_lock.head+1) % AG_MAXLOCK;
		if (gamma_mlock((buf),(size)) >= 0) {
AG_MEM_DEBUG(
printf("[%d] AG_mlock(): lock OK\n",
gamma_my_node());
fflush(stdout);
)
			return 0;
		}

		/*
		Lockable memory runs low!
		Try unlocking
		all the unlockable buffers...
		*/
AG_MEM_DEBUG(
printf("[%d] AG_mlock(): Trying unlocking all the unlockable buffers...\n",
gamma_my_node());
fflush(stdout);
)
		for (i=0;i<AG_lock.how_many;i++)
			if (!(AG_lock.table[i].in_use)) {
				gamma_munlock(
					AG_lock.table[i].addr,
					AG_lock.table[i].len
				);
				AG_lock.table[j].addr = NULL;
				AG_lock.table[j].len = 0;
			}
		/*  Now retry locking...  */
		if (gamma_mlock((buf),(size)) >= 0) {
AG_MEM_DEBUG(
printf("[%d] AG_mlock(): lock OK\n",
gamma_my_node());
fflush(stdout);
)
			return 0;
		}

		/*  mlock failed!  Bad news...  */
printf("[%d] AG_mlock(): out of lockable RAM!!!\n",
gamma_my_node());
fflush(stdout);
		return -1;
	}

	/*  First entry of lock table to be filled in.  */

	AG_lock.how_many++;
	AG_lock.table[AG_lock.head].addr = (buf);
	AG_lock.table[AG_lock.head].len = (size);
	AG_lock.table[AG_lock.head].in_use++;
AG_MEM_DEBUG_VERY_LIGHT(
printf("[%d] AG_mlock(): first lock OK, how_many %d, head %d, in_use %d\n",
gamma_my_node(), AG_lock.how_many,
(AG_lock.head+1)%AG_MAXLOCK,AG_lock.table[AG_lock.head].in_use);
fflush(stdout);
)
	AG_lock.head = (AG_lock.head + 1) % AG_MAXLOCK;
	if (gamma_mlock((buf),(size)) >= 0) return 0;

	/*  mlock failed!  Bad news...  */
printf("[%d] AG_mlock(): out of lockable RAM!!!\n",
gamma_my_node());
fflush(stdout);
	return -1;

}


void AG_memunlock(void *buf, unsigned long size) {
	int i,n;

	if (size == 0) return;
	n = AG_lock.how_many;
AG_MEM_DEBUG(
printf("[%d] AG_munlock(): how_many %d, head %d, start search at %d\n",
gamma_my_node(),AG_lock.how_many,AG_lock.head,
(AG_lock.head==0 ? AG_MAXLOCK-1 : AG_lock.head-1)
);
fflush(stdout);
)
	for (	i = (AG_lock.head==0 ? AG_MAXLOCK-1 : AG_lock.head-1);
		n;
		n--, i = (i==0 ? AG_MAXLOCK-1 : i-1)
	) {
		if (!(AG_lock.table[i].in_use)) continue;
		if (
(
(((unsigned long)buf)
>= (unsigned long)(AG_lock.table[i].addr))
&&
(((unsigned long)buf)
<= ((unsigned long)(AG_lock.table[i].addr)+AG_lock.table[i].len))
)
||
(
(((unsigned long)buf+size)
>= (unsigned long)(AG_lock.table[i].addr))
&&
(((unsigned long)buf+size)
<= ((unsigned long)(AG_lock.table[i].addr)+AG_lock.table[i].len))
)
		) {
			AG_lock.table[i].in_use--;
AG_MEM_DEBUG_LIGHT(
printf("[%d] AG_munlock(): found at %d, examined %d entries\n",
gamma_my_node(),i,AG_lock.how_many-n+1);
fflush(stdout);
)
			return;
		}
	}
printf("[%d] AG_munlock(): internal error!!\n",
gamma_my_node());
fflush(stdout);
	return;
}
#endif

#endif
